# ABSTRACT: A Magical Scrappy Namespace Where Wild Things Are
package Scrappy::Doo;
BEGIN {
  $Scrappy::Doo::VERSION = '0.61';
}
use Scrappy qw/:syntax/;
1;
__END__
=pod

=head1 NAME

Scrappy::Doo - A Magical Scrappy Namespace Where Wild Things Are

=head1 VERSION

version 0.61

=head1 AUTHOR

  Al Newkirk <awncorp@cpan.org>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2010 by awncorp.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut

